import json
import pathlib
import shutil
import tempfile
import uuid

import pytest

import dcoraid
import dcoraid.api
from dcoraid.api import dataset_create
from dcoraid.upload import job, task

import common


dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def test_custom_dataset_dict():
    api = common.get_api()
    # post dataset creation request
    task_path = common.make_upload_task(dataset_dict=False,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name])
    dataset_dict = common.make_dataset_dict()
    dataset_dict["authors"] = "Captain Hook!"
    uj = task.load_task(task_path, api=api,
                        dataset_kwargs=dataset_dict)
    # now make sure the authors were set correctly
    ddict = api.get("package_show", id=uj.dataset_id)
    assert ddict["authors"] == "Captain Hook!"


def test_custom_dataset_dict_2():
    api = common.get_api()
    # post dataset creation request
    task_path = common.make_upload_task(dataset_dict=True,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name])
    dataset_dict = common.make_dataset_dict()
    dataset_dict["authors"] = "Captain Hook!"
    uj = task.load_task(task_path, api=api,
                        dataset_kwargs=dataset_dict)
    # now make sure the authors were set correctly
    ddict = api.get("package_show", id=uj.dataset_id)
    assert ddict["authors"] == "Captain Hook!"


def test_dataset_id_already_exists_active_fails():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    # post dataset creation request
    dataset_dict_with_id = dataset_create(dataset_dict=dataset_dict,
                                          resources=[dpath],
                                          api=api,
                                          activate=True)
    # create a new task with the same dataset ID but with different data
    task_path = common.make_upload_task(
        dataset_dict=dataset_dict_with_id,
        resource_paths=[str(dpath), str(dpath)],
        resource_names=["1.rtdc", "2.rtdc"])
    uj = task.load_task(task_path, api=api)
    assert len(uj.paths) == 2
    assert len(uj.resource_names) == 2
    assert uj.dataset_id == dataset_dict_with_id["id"]
    # attempt to upload the task
    uj.task_compress_resources()
    assert uj.state == "parcel"
    with pytest.raises(dcoraid.api.APIAuthorizationError,
                       match=""):
        uj.task_upload_resources()


def test_dataset_id_does_not_exist():
    api = common.get_api()
    # create a fake ID
    dataset_id = str(uuid.uuid4())
    # create a new task with the fake dataset ID
    task_path = common.make_upload_task(dataset_id=dataset_id)
    # create the upload job
    with pytest.raises(dcoraid.api.APINotFoundError,
                       match=dataset_id):
        task.load_task(task_path, api=api)


def test_load_basic():
    api = common.get_api()
    task_path = common.make_upload_task(task_id="zpowiemsnh",
                                        resource_names=["humdinger.rtdc"])
    assert task.task_has_circle(task_path)
    uj = task.load_task(task_path, api=api)
    assert uj.task_id == "zpowiemsnh"
    assert uj.resource_names == ["humdinger.rtdc"]


def test_load_with_existing_dataset():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    # post dataset creation request
    dataset_dict_with_id = dataset_create(dataset_dict=dataset_dict,
                                          resources=[dpath],
                                          api=api)
    task_path = common.make_upload_task(dataset_dict=dataset_dict_with_id,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name])
    uj = task.load_task(task_path, api=api)
    assert uj.dataset_id == dataset_dict_with_id["id"]
    # skipping the upload should work, since it's already uploaded
    uj.set_state("online")
    uj.task_verify_resources()

    common.wait_for_job_no_queue(uj)


def test_load_with_existing_dataset_map_from_task():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    # post dataset creation request
    dataset_dict_with_id = dataset_create(dataset_dict=dataset_dict,
                                          resources=[dpath],
                                          api=api)
    task_path = common.make_upload_task(dataset_dict=dataset_dict,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name],
                                        task_id="xwing")
    uj = task.load_task(
        task_path,
        api=api,
        map_task_to_dataset_id={"xwing": dataset_dict_with_id["id"]})
    assert uj.dataset_id == dataset_dict_with_id["id"]


def test_load_with_existing_dataset_map_from_task_dict_update():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    # post dataset creation request
    task_path = common.make_upload_task(dataset_dict=dataset_dict,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name],
                                        task_id="xwing")
    map_task_to_dataset_id = {}
    uj = task.load_task(
        task_path,
        api=api,
        map_task_to_dataset_id=map_task_to_dataset_id)
    assert uj.task_id == "xwing"
    assert map_task_to_dataset_id["xwing"] == uj.dataset_id


def test_load_with_existing_dataset_map_from_task_control():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    # post dataset creation request
    dataset_dict_with_id = dataset_create(dataset_dict=dataset_dict,
                                          resources=[dpath],
                                          api=api)
    task_path = common.make_upload_task(dataset_dict=dataset_dict,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name],
                                        task_id="xwing")
    uj = task.load_task(
        task_path,
        api=api,
        map_task_to_dataset_id={"deathstar": dataset_dict_with_id["id"]})
    assert uj.dataset_id != dataset_dict_with_id["id"]


def test_load_with_update():
    api = common.get_api()
    task_path = common.make_upload_task(task_id="blackfalcon",
                                        resource_names=["marvel.rtdc"])
    assert task.task_has_circle(task_path)
    uj = task.load_task(task_path, api=api, update_dataset_id=True)
    assert uj.task_id == "blackfalcon"
    assert uj.resource_names == ["marvel.rtdc"]
    # Load task and check dataset_id
    with open(task_path) as fd:
        task_dict = json.load(fd)
        assert task_dict["dataset_dict"]["id"] == uj.dataset_id


def test_missing_owner_org():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    dataset_dict.pop("owner_org")
    task_path = common.make_upload_task(dataset_dict=dataset_dict)
    assert not task.task_has_circle(task_path)
    with pytest.raises(dcoraid.api.APIConflictError,
                       match="A circle must be provided"):
        task.load_task(task_path, api=api)


def test_no_ids():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    task_path = common.make_upload_task(dataset_dict=dataset_dict,
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name],
                                        task_id=None)
    with pytest.raises(ValueError,
                       match="or pass the dataset_id via the dataset_kwargs"):
        task.load_task(task_path, api=api)


def test_persistent_dict():
    tdir = tempfile.mkdtemp(prefix="upload_task_persistent_dict_")
    path = pathlib.Path(tdir) / "persistent_dict.txt"
    pd = task.PersistentTaskDatasetIDDict(path)
    pd["hans"] = "peter"
    pd["golem"] = "odin123"
    assert pd["hans"] == "peter"
    assert pd["golem"] == "odin123"
    assert pd.get("fritz") is None

    pd2 = task.PersistentTaskDatasetIDDict(path)
    assert pd2["hans"] == "peter"
    assert pd2["golem"] == "odin123"


def test_persistent_dict_bad_task_id():
    tdir = tempfile.mkdtemp(prefix="upload_task_persistent_dict_")
    path = pathlib.Path(tdir) / "persistent_dict.txt"
    pd = task.PersistentTaskDatasetIDDict(path)
    with pytest.raises(ValueError, match="task IDs may only contain"):
        pd["hans!"] = "peter"
    # make sure it is not stored
    pd2 = task.PersistentTaskDatasetIDDict(path)
    assert "hans!" not in pd2


def test_persistent_dict_override_forbidden():
    tdir = tempfile.mkdtemp(prefix="upload_task_persistent_dict_")
    path = pathlib.Path(tdir) / "persistent_dict.txt"
    pd = task.PersistentTaskDatasetIDDict(path)
    pd["peter"] = "pan"
    assert "peter" in pd
    assert pd["peter"] == "pan"
    pd["peter"] = "pan"  # once again
    with pytest.raises(ValueError, match="Cannot override entries"):
        pd["peter"] = "hook"
    # make sure it is not stored
    pd2 = task.PersistentTaskDatasetIDDict(path)
    assert pd2["peter"] == "pan"


def test_resource_name_lengths():
    """Make sure ValueError is raised when list lengths do not match"""
    task_path = common.make_upload_task(
        resource_paths=[__file__, dpath],
        resource_names=["other_data.rtdc"],
        resource_supplements=[{},
                              {"chip": {"name": "7x2",
                                        "master name": "R1"}}])
    with pytest.raises(ValueError,
                       match="does not match number of resource names"):
        task.load_task(task_path, api=common.get_api())


def test_resource_path_is_relative():
    task_path = common.make_upload_task(resource_paths=["guess_my_name.rtdc"])
    new_data_path = pathlib.Path(task_path).parent / "guess_my_name.rtdc"
    shutil.copy2(dpath, new_data_path)
    uj = task.load_task(task_path, api=common.get_api())
    assert new_data_path.samefile(uj.paths[0])


def test_resource_path_not_found():
    task_path = common.make_upload_task(resource_paths=["/home/unknown.rtdc"])
    with pytest.raises(task.LocalTaskResourcesNotFoundError,
                       match="is missing local resources files"):
        task.load_task(task_path, api=common.get_api())


def test_resource_supplements():
    task_path = common.make_upload_task(
        resource_paths=[dpath],
        resource_supplements=[{"chip": {"name": "7x2",
                                        "master name": "R1"}}])
    uj = task.load_task(task_path, api=common.get_api())
    assert uj.supplements[0]["chip"]["name"] == "7x2"
    assert uj.supplements[0]["chip"]["master name"] == "R1"


def test_resource_supplements_must_be_empty_for_non_rtdc():
    task_path = common.make_upload_task(
        resource_paths=[__file__, dpath],
        resource_names=["test.py", "other_data.rtdc"],
        resource_supplements=[{"chip": {"name": "7x2",
                                        "master name": "R1"}},
                              {"chip": {"name": "7x2",
                                        "master name": "R1"}}
                              ])
    with pytest.raises(ValueError, match="supplements must be empty"):
        task.load_task(task_path, api=common.get_api())


def test_resource_supplements_with_other_files():
    task_path = common.make_upload_task(
        resource_paths=[__file__, dpath],
        resource_names=["test.py", "other_data.rtdc"],
        resource_supplements=[{},
                              {"chip": {"name": "7x2",
                                        "master name": "R1"}}])
    uj = task.load_task(task_path, api=common.get_api())
    assert len(uj.supplements[0]) == 0


def test_resource_supplements_lengths():
    """Make sure ValueError is raised when list lengths do not match"""
    task_path = common.make_upload_task(
        resource_paths=[__file__, dpath],
        resource_names=["test.py", "other_data.rtdc"],
        resource_supplements=[{"chip": {"name": "7x2",
                                        "master name": "R1"}}])
    with pytest.raises(ValueError,
                       match="does not match number of resource supplements"):
        task.load_task(task_path, api=common.get_api())


def test_save_load():
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=[dpath], task_id="hanspeter")
    td = pathlib.Path(tempfile.mkdtemp(prefix="task_"))
    task_path = td / "test.dcoraid-task"
    task.save_task(uj, path=task_path)
    uj2 = task.load_task(task_path, api=api)
    assert uj.dataset_id == uj2.dataset_id
    assert uj.paths[0].samefile(uj2.paths[0])


def test_wrong_ids():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="task_test")
    dataset_dict["id"] = "peter"
    task_path = common.make_upload_task(dataset_dict=dataset_dict,
                                        dataset_id="hans",  # different id
                                        resource_paths=[str(dpath)],
                                        resource_names=[dpath.name])
    with pytest.raises(ValueError,
                       match="I got the following IDs: from upload job "
                             + "state: hans; from dataset dict: peter"):
        task.load_task(task_path, api=api)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import urllib.parse

import numpy as np

from ..common import ttl_cache

from .db_core import DBInterrogator
from .extract import DBExtract


class APIInterrogator(DBInterrogator):
    def __init__(self, api):
        self.api = api.copy()
        if api.user_id:
            mode = "user"
            user_data = api.get_user_dict()
        else:
            mode = "public"
            user_data = None
        super(APIInterrogator, self).__init__(mode=mode, user_data=user_data)

    @ttl_cache(seconds=5)
    def get_circles(self):
        """Return the list of DCOR Circle names
        """
        if self.mode == "user":
            # Organizations the user is a member of
            circle_dict = self.api.get("organization_list_for_user",
                                       id=self.api.user_id,
                                       permission="read")
            data = [dd["name"] for dd in circle_dict]
        else:
            data = self.api.get("organization_list")
        return data

    @ttl_cache(seconds=5)
    def get_collections(self):
        """Return the list of DCOR Collection names"""
        if self.mode == "user":
            collection_dict = self.api.get("group_list_authz", am_member=True)
            data = [dd["name"] for dd in collection_dict]
        else:
            data = self.api.get("group_list")
        if len(data) == 1000:
            raise NotImplementedError(
                "Reached hard limit of 1000 results! "
                + "Please ask someone to implement this with `offset`.")
        return data

    @ttl_cache(seconds=3600)
    def get_datasets_user_following(self):
        """Return datasets the user is following"""
        assert self.mode == "user"
        data = self.api.get("dataset_followee_list", id=self.api.user_name)
        return DBExtract(data)

    @ttl_cache(seconds=3600)
    def get_datasets_user_owned(self):
        """Return datasets the user created"""
        assert self.mode == "user"
        datasets = self.search_dataset(
            filter_queries=[f"+creator_user_id:{self.api.user_id}"],
            limit=0,
        )
        return datasets

    def get_datasets_user_shared(self):
        """Return datasets shared with the user"""
        assert self.mode == "user"
        # TODO:
        # - package_collaborator_list_for_user
        #   - https://github.com/DCOR-dev/DCOR-Aid/issues/32
        #   - https://github.com/DCOR-dev/ckanext-dcor_schemas/issues/10

        # perform a dataset search with those circles and collections
        datasets = self.search_dataset(
            circles=self.get_circles(),
            collections=self.get_collections(),
            circle_collection_union=True,
            filter_queries=[f"-creator_user_id:{self.api.user_id}"],
            limit=0,
        )
        return datasets

    @ttl_cache(seconds=3600)
    def get_users(self, ret_fullnames=False):
        """Return the list of DCOR users"""
        data = self.api.get("user_list")
        user_list = []
        full_list = []
        for dd in data:
            user_list.append(dd["name"])
            if dd["fullname"]:
                full_list.append(dd["fullname"])
            else:
                full_list.append(dd["name"])
        if ret_fullnames:
            return user_list, full_list
        else:
            return user_list

    def search_dataset(self, query="*:*", filter_queries=None, circles=None,
                       collections=None, circle_collection_union=False,
                       limit=100):
        """Search datasets via the CKAN API

        Parameters
        ----------
        query: str
            search query
        filter_queries: list of str
            SOLR `fq` filter queries (are joined with 'AND')
        circles: list of str
            list of circles (organizations) to search in
        collections: list of str
            list of collections (groups) to search in
        circle_collection_union: bool
            If set to True, make a union of the circle and collection
            sets. Otherwise (default), search only for datasets that
            are are at least member of one of the circles and one of the
            collections.
        limit: int
            limit number of search results; Set to 0 to get all results
        """
        if filter_queries is None:
            filter_queries = []
        # https://docs.ckan.org/en/latest/user-guide.html#search-in-detail
        if circles:
            solr_circles = ["organization:{}".format(ci) for ci in circles]
            if len(circles) == 1:
                solr_circle_query = solr_circles[0]
            else:
                solr_circle_query = f"({' OR '.join(solr_circles)})"
        else:
            solr_circle_query = None

        if collections:
            solr_collections = ["groups:{}".format(co) for co in collections]
            if len(collections) == 1:
                solr_collections_query = solr_collections[0]
            else:
                solr_collections_query = f"({' OR '.join(solr_collections)})"
        else:
            solr_collections_query = None

        if solr_circle_query and solr_collections_query:
            if circle_collection_union:
                fq = f"({solr_circle_query} OR {solr_collections_query})"
            else:
                fq = f"({solr_circle_query} AND {solr_collections_query})"
        elif solr_circle_query:
            fq = f"{solr_circle_query}"
        elif solr_collections_query:
            fq = f"{solr_collections_query}"
        else:
            fq = ""
        if fq:
            filter_queries.append(fq)
        if len(filter_queries) == 0:
            final_fq = ""
        elif len(filter_queries) == 1:
            final_fq = filter_queries[0]
        else:
            final_fq = f"({' AND '.join(filter_queries)})"

        if limit < 0:
            raise ValueError(f"`limit` must be 0 or >0!, got {limit}!")
        elif limit == 0:
            rows = 1000  # default batch size
            limit = np.inf
        else:
            rows = min(1000, limit)

        num_total = np.inf  # just the initial value
        num_retrieved = 0
        dbextract = DBExtract()
        while num_retrieved < min(limit, num_total) and rows:
            data = self.api.get("package_search",
                                q=urllib.parse.quote(query, safe=""),
                                fq=final_fq,
                                include_private=(self.mode == "user"),
                                rows=rows,
                                start=num_retrieved,
                                )
            if np.isinf(num_total):
                # first iteration
                num_total = data["count"]
            num_retrieved += len(data["results"])
            if num_retrieved + rows > min(limit, num_total):
                # in the next iteration, only get the final
                # few results.
                rows = num_total - num_retrieved
            dbextract.add_datasets(data["results"])

        return dbextract

    @property
    def local_timestamp(self):
        """Local database date in seconds since epoch"""
        return self.remote_timestamp

    @property
    def local_version_score(self):
        """Local database version"""
        return self.remote_version_score



import pathlib
import shutil
import tempfile
from unittest import mock

import uuid

from dcoraid.gui.main import DCORAid
from dcoraid.gui.upload.dlg_upload import UploadDialog
from dcoraid.gui.upload import widget_upload

import pytest
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QMessageBox

import common


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    pass
    # Run test
    yield
    # Make sure that all daemons are gone
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                         3000)
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                         3000)


@pytest.mark.filterwarnings("ignore::UserWarning", match="No API key is set!")
def test_anonymous(qtbot):
    """Start DCOR-Aid in anonymous mode"""
    QtCore.QCoreApplication.setOrganizationName("DCOR")
    QtCore.QCoreApplication.setOrganizationDomain("dcor.mpl.mpg.de")
    QtCore.QCoreApplication.setApplicationName("dcoraid")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    spath = pathlib.Path(settings.fileName())
    # temporarily move settings to temporary location
    stmp = pathlib.Path(tempfile.mkdtemp(prefix="settings_stash_")) / "set.ini"
    shutil.copy2(spath, stmp)
    spath.unlink()
    spath.write_text("\n".join([
        "[General]",
        "user%20scenario = anonymous",
        "[auth]",
        "api%20key =",
        "server = dcor.mpl.mpg.de",
        ]))
    try:
        mw = DCORAid()
        QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)
        # sanity check
        assert mw.settings.value("user scenario") == "anonymous"
        assert mw.settings.value("auth/server") == "dcor.mpl.mpg.de"
    except BaseException:
        # cleanup first (copy back original settings for other tests)
        spath.unlink()
        shutil.copy2(stmp, spath)
        raise
    else:
        spath.unlink()
        shutil.copy2(stmp, spath)
    mw.close()


def test_upload_simple(qtbot, monkeypatch):
    """Upload a test dataset"""
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)

    dlg = UploadDialog(mw.panel_upload)
    mw.panel_upload._dlg_manual = dlg
    dlg.finished.connect(mw.panel_upload.on_upload_manual_ready)
    # Fill data for testing
    dlg._autofill_for_testing()
    # Avoid message boxes
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    # Commence upload
    dlg.on_proceed()
    assert dlg.dataset_id is not None

    common.wait_for_job(upload_queue=mw.panel_upload.jobs,
                        dataset_id=mw.panel_upload.jobs[0].dataset_id)
    mw.close()


def test_upload_task(qtbot, monkeypatch):
    task_id = str(uuid.uuid4())
    tpath = common.make_upload_task(task_id=task_id)
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)
    monkeypatch.setattr(QtWidgets.QFileDialog, "getOpenFileNames",
                        lambda *args: ([tpath], None))
    act = QtWidgets.QAction("some unimportant text")
    act.setData("single")
    mw.panel_upload.on_upload_task(action=act)
    uj = mw.panel_upload.jobs[-1]
    assert uj.task_id == task_id
    mw.close()


def test_upload_task_missing_circle(qtbot, monkeypatch):
    """When the organization is missing, DCOR-Aid should ask for it"""
    task_id = str(uuid.uuid4())
    dataset_dict = common.make_dataset_dict(hint="task_upload_no_org_")
    dataset_dict.pop("owner_org")
    tpath = common.make_upload_task(task_id=task_id,
                                    dataset_dict=dataset_dict)
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)
    monkeypatch.setattr(QtWidgets.QFileDialog, "getOpenFileNames",
                        lambda *args: ([tpath], None))
    # We actually only need this monkeypatch if there is more than
    # one circle for the present user.
    monkeypatch.setattr(QtWidgets.QInputDialog, "getItem",
                        # return the first item in the circle list
                        lambda *args: (args[3][0], True))
    act = QtWidgets.QAction("some unimportant text")
    act.setData("single")
    mw.panel_upload.on_upload_task(action=act)
    uj = mw.panel_upload.jobs[-1]
    assert uj.task_id == task_id
    mw.close()


def test_upload_task_missing_circle_multiple(qtbot, monkeypatch):
    """DCOR-Aid should only ask *once* for the circle (not for every task)"""
    task_id1 = str(uuid.uuid4())
    dataset_dict1 = common.make_dataset_dict(hint="task_upload_no_org_")
    dataset_dict1.pop("owner_org")
    tpath1 = common.make_upload_task(task_id=task_id1,
                                     dataset_dict=dataset_dict1)
    tpath1 = pathlib.Path(tpath1)

    task_id2 = str(uuid.uuid4())
    dataset_dict2 = common.make_dataset_dict(hint="task_upload_no_org_")
    dataset_dict2.pop("owner_org")
    tpath2 = common.make_upload_task(task_id=task_id2,
                                     dataset_dict=dataset_dict2)
    tpath2 = pathlib.Path(tpath2)

    tdir = pathlib.Path(tempfile.mkdtemp(prefix="recursive_task_"))
    shutil.copytree(tpath1.parent, tdir / tpath1.parent.name)
    shutil.copytree(tpath2.parent, tdir / tpath2.parent.name)

    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)
    monkeypatch.setattr(QtWidgets.QFileDialog, "getExistingDirectory",
                        lambda *args: str(tdir))
    # We actually only need this monkeypatch if there is more than
    # one circle for the present user.
    monkeypatch.setattr(QtWidgets.QInputDialog, "getItem",
                        # return the first item in the circle list
                        lambda *args: (args[3][0], True))
    act = QtWidgets.QAction("some unimportant text")
    act.setData("bulk")
    request_circle = widget_upload.circle_mgr.request_circle
    with mock.patch.object(widget_upload.circle_mgr,
                           "request_circle",
                           wraps=request_circle) as rw:
        mw.panel_upload.on_upload_task(action=act)
        uj1 = mw.panel_upload.jobs[-2]
        uj2 = mw.panel_upload.jobs[-1]
        assert {uj1.task_id, uj2.task_id} == {task_id1, task_id2}
        assert rw.call_count == 1
    mw.close()


def test_upload_private(qtbot, monkeypatch):
    """Upload a private test dataset"""
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)

    dlg = UploadDialog(mw.panel_upload)
    mw.panel_upload._dlg_manual = dlg
    dlg.finished.connect(mw.panel_upload.on_upload_manual_ready)
    # Fill data for testing
    dlg._autofill_for_testing()
    # set visibility to private
    dlg.comboBox_vis.setCurrentIndex(dlg.comboBox_vis.findData("private"))
    # Avoid message boxes
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    # Commence upload
    dlg.on_proceed()
    dataset_id = dlg.dataset_id
    assert dataset_id is not None

    common.wait_for_job(upload_queue=mw.panel_upload.jobs,
                        dataset_id=mw.panel_upload.jobs[0].dataset_id)
    mw.close()
    # make sure the dataset is private
    api = common.get_api()
    dataset_dict = api.get(api_call="package_show", id=dataset_id)
    assert dataset_dict["private"]
    assert isinstance(dataset_dict["private"], bool)



import functools
import hashlib
import pathlib
import time
from functools import lru_cache

import requests


ConnectionTimeoutErrors = (ConnectionError,
                           requests.exceptions.ConnectionError,
                           requests.exceptions.Timeout)


@lru_cache(maxsize=2000)
def sha256sum(path):
    """Compute the SHA256 sum of a file on disk"""
    block_size = 2**20
    path = pathlib.Path(path)
    file_hash = hashlib.sha256()
    with path.open("rb") as fd:
        while True:
            data = fd.read(block_size)
            if not data:
                break
            file_hash.update(data)
    return file_hash.hexdigest()


def ttl_cache(seconds=5, maxsize=32, typed=False):
    """A time-to-live cache based on lru_cache"""
    def wrapper_cache(func):
        func = functools.lru_cache(maxsize=maxsize, typed=typed)(func)
        func.tinit = time.perf_counter()
        func.delta = 0

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            delta = (time.perf_counter() - func.tinit) // seconds
            if delta != func.delta:
                func.cache_clear()
                func.delta = delta
            return func(*args, **kwargs)
        return wrapped_func

    return wrapper_cache



from functools import lru_cache

from PyQt5 import QtWidgets

from ..api import get_ckan_api


@lru_cache(maxsize=1)
def get_user_circle_dicts():
    """Conveniently cache the organization list for the user

    You must call `self._get_user_circle_dicts.cache_clear`
    if the circle list changes (e.g. because you added one).
    """
    api = get_ckan_api()
    circles = api.get("organization_list_for_user",
                      permission="create_dataset")
    return circles


def ask_for_new_circle(parent_widget):
    """Ask the user to create a new circle

    Returns
    -------
    circle_dict: dict or None
        The CKAN organization dictionary. None is returned if the
        user did not create a circle.
    """
    api = get_ckan_api()
    # Ask the user whether he would like to create a circle.
    ud = api.get_user_dict()
    name = ud["fullname"] if ud["fullname"] else ud["name"]
    text, ok_pressed = QtWidgets.QInputDialog.getText(
        parent_widget,
        "Circle required",
        "You do not have access to any existing Circles. To upload\n"
        + "datasets, you need to be either Editor or Admin in\n"
        + "a Circle. You may create a Circle now or cancel and ask\n"
        + "a colleague to add you to a Circle (Your user name is "
        + "'{}').".format(ud["name"])
        + "\n\nTo proceed with Circle creation, please choose a name:",
        QtWidgets.QLineEdit.Normal,
        "{}'s Circle".format(name))
    if ok_pressed and text != '':
        cname = "user-circle-{}".format(ud["name"])
        cdict = api.post("organization_create",
                         data={"name": cname,
                               "title": text.strip(),
                               })
        # invalidate cache, because now we have a new circle
        get_user_circle_dicts.cache_clear()
        return cdict


def request_circle(parent_widget):
    """Ask the user to select a circle

    If only a single circle is present in the list, than this circle
    is used without user interaction.

    Returns
    -------
    circle_dict: dict or None
        The CKAN organization dictionary. None is returned if the
        user aborted.
    """
    circs = get_user_circle_dicts()
    if len(circs) == 1:
        return circs[0]
    elif len(circs) == 0:
        return ask_for_new_circle(parent_widget)
    else:
        # Show a dialog with choices
        circle_texts = [c["title"] if c["title"] else c["name"] for c in circs]
        text, ok_pressed = QtWidgets.QInputDialog.getItem(
            parent_widget,
            "Choose circe",
            "You have upload-permissions to multiple circles. Please\n"
            "choose the one where you wish to upload your datasets to.",
            circle_texts,
            0,  # default index
            False,  # not editable
        )
        if ok_pressed:
            cid = circle_texts.index(text)
            return circs[cid]



import copy
import pkg_resources

from PyQt5 import QtCore, QtGui, QtWidgets, uic


class FilterBase(QtWidgets.QWidget):
    #: The user selection has changed
    selection_changed = QtCore.pyqtSignal(list)
    download_resource = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        """Filter view widget with title, edit, checkbox, and table
        """
        super(FilterBase, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("dcoraid.gui.dbview",
                                                  "filter_base.ui")
        uic.loadUi(path_ui, self)

        #: List of entries in the current list
        self.entries = []

        # resize first column
        header = self.tableWidget.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

        # trigger user selection change signal
        self.tableWidget.itemSelectionChanged.connect(self.on_entry_selected)

        # TODO: enable quick-filters via lineEdit
        self.lineEdit.setVisible(False)

    def get_entry_actions(self, row, entry):
        """This is defined in the subclasses (Circle, Collection, etc)"""
        return []

    def get_entry_identifiers(self, selected=False):
        """Return the identifiers of the current tableWidget entries"""
        if selected:
            identifiers = []
            for ii, entry in enumerate(self.entries):
                if self.tableWidget.item(ii, 0).isSelected():
                    identifiers.append(entry["name"])
        else:
            identifiers = [ee["name"] for ee in self.entries]
        return identifiers

    @QtCore.pyqtSlot()
    def on_entry_selected(self):
        ids = self.get_entry_identifiers(selected=True)
        self.selection_changed.emit(ids)

    def set_entries(self, entries):
        """Set the current tableWidget entries

        """
        if not isinstance(entries, list):
            raise ValueError(f"`entries` must be list, got '{entries}'!")
        self.tableWidget.clear()
        self.entries = copy.deepcopy(entries)
        self.tableWidget.blockSignals(True)
        self.tableWidget.setRowCount(len(self.entries))
        for row, entry in enumerate(self.entries):
            self.set_entry(row, entry)
        self.tableWidget.blockSignals(False)

    def set_entry(self, row, entry):
        """Set table Widget entry at index `row`"""
        # text (1st column)
        self.set_entry_label(row, entry)

        # tool buttons (2nd column)
        widact = QtWidgets.QWidget(self)
        horz_layout = QtWidgets.QHBoxLayout(widact)
        horz_layout.setContentsMargins(2, 0, 2, 0)

        spacer = QtWidgets.QSpacerItem(0, 0,
                                       QtWidgets.QSizePolicy.Expanding,
                                       QtWidgets.QSizePolicy.Minimum)
        horz_layout.addItem(spacer)

        for action in self.get_entry_actions(row, entry):
            tbact = QtWidgets.QToolButton(widact)
            icon = QtGui.QIcon.fromTheme(action["icon"])
            tbact.setIcon(icon)
            tbact.setToolTip(action["tooltip"])
            tbact.clicked.connect(action["function"])
            horz_layout.addWidget(tbact)
        self.tableWidget.setCellWidget(row, 1, widact)

    def set_entry_label(self, row, entry):
        """Set table Widget entry text at index `row`

        Parameters
        ----------
        row: int
            row where to put the entry
        entry: dict
            CKAN entry dictionary
        """
        item = QtWidgets.QTableWidgetItem()
        item.setText(entry.get("title") or entry["name"])
        self.tableWidget.setItem(row, 0, item)



from PyQt5 import QtWidgets, QtCore, QtGui
from PyQt5.QtCore import Qt


class DragTableWidget(QtWidgets.QTableWidget):
    def mouseMoveEvent(self, e):
        if e.buttons() != Qt.LeftButton:
            return

        urls = []
        for item in self.selectedItems():
            data = item.data(Qt.UserRole + 1)
            urls.append(QtCore.QUrl(data))

        mime_data = QtCore.QMimeData()
        mime_data.setUrls(urls)

        drag = QtGui.QDrag(self)
        drag.setMimeData(mime_data)
        drag.setHotSpot(e.pos() - self.rect().topLeft())

        # This magic is somehow required to get drag working.
        dropAction = drag.exec_(Qt.CopyAction)  # noqa: F841
